# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
gcc-toolchain compilation option related helper functions.
"""


import os
import re


def toolchain_in_args(compiler_option):
    """
    Check for the --gcc-toolchain in the compilation options.

    """
    for cmp_opt in compiler_option:
        if '--gcc-toolchain' in cmp_opt:
            tcpath = \
                re.match(r"^--gcc-toolchain=(?P<tcpath>.*)$",
                         cmp_opt).group('tcpath')
            return tcpath

    return None


def get_toolchain_compiler(toolchain_path, language):
    """
    The compiler binary is expected to be under the bin directory
    in the tooolchain path.

    Construct the binary path base on the compilation language.
    """

    if language == 'c':
        return os.path.join(toolchain_path, 'bin', 'gcc')
    elif language == 'c++':
        return os.path.join(toolchain_path, 'bin', 'g++')

    return None



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Helper for tha authentication api.
"""

from codechecker_api.Authentication_v6 import codeCheckerAuthentication

from codechecker_client.thrift_call import ThriftClientCall
from .base import BaseClientHelper


class ThriftAuthHelper(BaseClientHelper):
    def __init__(self, protocol, host, port, uri, session_token=None):
        super().__init__(protocol, host, port, uri, session_token)

        self.client = codeCheckerAuthentication.Client(self.protocol)

    @ThriftClientCall
    def checkAPIVersion(self):
        pass

    # ============= Authentication and session handling =============
    @ThriftClientCall
    def getAuthParameters(self):
        pass

    @ThriftClientCall
    def getAcceptedAuthMethods(self):
        pass

    @ThriftClientCall
    def getAccessControl(self):
        pass

    @ThriftClientCall
    def performLogin(self, auth_method, auth_string):
        pass

    @ThriftClientCall
    def destroySession(self):
        pass

    # ============= Authorization, permission management =============
    @ThriftClientCall
    def getPermissions(self, scope):
        pass

    @ThriftClientCall
    def getPermissionsForUser(self, scope, extra_params, filter):
        pass

    @ThriftClientCall
    def getAuthorisedNames(self, permission, extra_params):
        pass

    @ThriftClientCall
    def addPermission(self, permission, auth_name, is_group, extra_params):
        pass

    @ThriftClientCall
    def removePermission(self, permission, auth_name, is_group, extra_params):
        pass

    @ThriftClientCall
    def hasPermission(self, permission, extra_params):
        pass

    # ============= Token management =============

    @ThriftClientCall
    def newToken(self, description):
        pass

    @ThriftClientCall
    def removeToken(self, token):
        pass

    @ThriftClientCall
    def getTokens(self):
        pass



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Context to store package related information.
"""


from collections.abc import Mapping
import os
import re
import sys

from codechecker_common import logger
from codechecker_common.checker_labels import CheckerLabels
from codechecker_common.singleton import Singleton
from codechecker_common.util import load_json

LOG = logger.get_logger('system')


class SeverityMap(Mapping):
    """
    A dictionary which maps checker names to severity levels.
    If a key is not found in the map then it will return MEDIUM severity in
    case of compiler warnings and CRITICAL in case of compiler errors.
    """

    def __init__(self, *args, **kwargs):
        self.store = dict(*args, **kwargs)

    def __getitem__(self, key):
        # Key is not specified in the store and it is a compiler warning
        # or error.
        if key not in self.store:
            if key == 'clang-diagnostic-error':
                return "CRITICAL"
            elif key.startswith('clang-diagnostic-'):
                return "MEDIUM"

        return self.store[key] if key in self.store else 'UNSPECIFIED'

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)


# -----------------------------------------------------------------------------
class Context(metaclass=Singleton):
    """ Generic package specific context. """

    def __init__(self):
        """ Initialize web context. """
        self._lib_dir_path = os.environ.get('CC_LIB_DIR', '')
        self._data_files_dir_path = os.environ.get('CC_DATA_FILES_DIR', '')

        lcfg_dict = self.__get_package_layout()
        self.pckg_layout = lcfg_dict['runtime']

        # Use this environment variable for testing purposes only. This
        # variable helps to configure which labels to use in this context.
        labels_dir = os.path.join(self._data_files_dir_path,
                                  'config', 'labels')
        if 'CC_TEST_LABELS_DIR' in os.environ:
            labels_dir = os.environ['CC_TEST_LABELS_DIR']

        self._checker_labels = CheckerLabels(labels_dir)
        self.__system_comment_map = load_json(self.system_comment_map_file, {})
        self.__git_commit_urls = self.__get_git_commit_urls()
        self.__package_version = None
        self.__package_build_date = None
        self.__package_git_hash = None

        # This should be initialized in command line scripts based on the
        # given CLI options.
        self.codechecker_workspace = None

        self.__set_version()

    def __get_git_commit_urls(self):
        """ Get commit urls from the configuration file. """
        git_commit_urls = load_json(self.git_commit_urls_file, [])

        for git_commit_url in git_commit_urls:
            git_commit_url["regex"] = re.compile(git_commit_url["regex"])

        return git_commit_urls

    def __get_package_layout(self):
        """ Get package layout configuration. """
        layout_cfg_file = os.path.join(
            self._data_files_dir_path, "config", "package_layout.json")

        LOG.debug('Reading config: %s', layout_cfg_file)
        lcfg_dict = load_json(layout_cfg_file)

        if not lcfg_dict:
            raise ValueError(f"No configuration file '{layout_cfg_file}' can "
                             f"be found or it is empty!")

        return lcfg_dict

    def __set_version(self):
        """
        Get the package version from the version config file.
        """
        vfile_data = load_json(self.version_file)

        if not vfile_data:
            sys.exit(1)

        package_version = vfile_data['version']
        package_build_date = vfile_data['package_build_date']
        package_git_hash = vfile_data.get('git_hash')
        package_git_tag = vfile_data.get('git_describe', {}).get('tag')
        package_git_dirtytag = vfile_data.get('git_describe', {}).get('dirty')

        self.__package_version = package_version['major'] + '.' + \
            package_version['minor'] + '.' + \
            package_version['revision'] + \
            ('-rc' + package_version['rc'] if 'rc' in package_version and
             package_version['rc'] else '')

        self.__package_build_date = package_build_date
        self.__package_git_hash = package_git_hash

        self.__package_git_tag = package_git_tag
        if (LOG.getEffectiveLevel() == logger.DEBUG or
                LOG.getEffectiveLevel() ==
                logger.DEBUG_ANALYZER):
            self.__package_git_tag = package_git_dirtytag

    @property
    def version(self):
        return self.__package_version

    @property
    def package_build_date(self):
        return self.__package_build_date

    @property
    def package_git_hash(self):
        return self.__package_git_hash

    @property
    def package_git_tag(self):
        return self.__package_git_tag

    @property
    def version_file(self):
        return os.path.join(self._data_files_dir_path, 'config',
                            'web_version.json')

    @property
    def system_comment_map(self):
        return self.__system_comment_map

    @property
    def system_comment_map_file(self):
        return os.path.join(self._data_files_dir_path, 'config',
                            'system_comment_kinds.json')

    @property
    def git_commit_urls_file(self):
        return os.path.join(
            self._data_files_dir_path, 'config', 'git_commit_urls.json')

    @property
    def git_commit_urls(self):
        return self.__git_commit_urls

    @property
    def path_plist_to_html_dist(self):
        return os.path.join(self._lib_dir_path, 'codechecker_report_converter',
                            'report', 'output', 'html', 'static')

    @property
    def path_env_extra(self):
        extra_paths = self.pckg_layout.get('path_env_extra', [])
        paths = []
        for path in extra_paths:
            paths.append(os.path.join(self._data_files_dir_path, path))
        return paths

    @property
    def ld_lib_path_extra(self):
        extra_lib = self.pckg_layout.get('ld_lib_path_extra', [])
        ld_paths = []
        for path in extra_lib:
            ld_paths.append(os.path.join(self._data_files_dir_path, path))
        return ld_paths

    @property
    def data_files_dir_path(self):
        return self._data_files_dir_path

    @property
    def doc_root(self):
        return os.path.join(self._data_files_dir_path, 'www', 'docs')

    @property
    def www_root(self):
        return os.path.join(self._data_files_dir_path, 'www')

    @property
    def run_migration_root(self):
        return os.path.join(self._lib_dir_path, 'codechecker_server',
                            'migrations', 'report')

    @property
    def config_migration_root(self):
        return os.path.join(self._lib_dir_path, 'codechecker_server',
                            'migrations', 'config')

    @property
    def checker_labels(self):
        return self._checker_labels


def get_context():
    try:
        return Context()
    except KeyError:
        import traceback
        traceback.print_exc()
        sys.exit(1)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

from typing import Dict


class Singleton(type):
    """ Helper type to create singleton classes. """

    _instances: Dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = \
                super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Thrift client setup and configuration.
"""


import getpass
import sys

from thrift.Thrift import TApplicationException

import codechecker_api_shared
from codechecker_api.Authentication_v6 import ttypes as AuthTypes

from codechecker_common.logger import get_logger

from codechecker_web.shared import env
from codechecker_web.shared.version import CLIENT_API

from codechecker_client.helpers.authentication import ThriftAuthHelper
from codechecker_client.helpers.product import ThriftProductHelper
from codechecker_client.helpers.results import ThriftResultsHelper
from .credential_manager import UserCredentials
from .product import split_product_url

LOG = get_logger('system')


def check_preconfigured_username(username, host, port):
    """
    Checks if username supplied by using preconfigured credentials.
    """
    if not username:
        LOG.error("No username supplied! Please specify the "
                  "username in your \"%s\" file for %s:%d.",
                  env.get_password_file(), host, port)
        sys.exit(1)


def init_auth_client(protocol, host, port):
    """ Setup a new auth client. """
    auth_client = setup_auth_client(protocol, host, port)

    # Check if local token is available.
    cred_manager = UserCredentials()
    session_token = cred_manager.get_token(host, port)

    if not session_token:
        LOG.info("No valid token or session was found for %s:%s", host, port)
        session_token = perform_auth_for_handler(auth_client, host, port,
                                                 cred_manager)

    return setup_auth_client(protocol, host, port, session_token)


def setup_auth_client(protocol, host, port, session_token=None):
    """
    Setup the Thrift authentication client. Returns the client object and the
    session token for the session.
    """
    client = ThriftAuthHelper(protocol, host, port,
                              '/v' + CLIENT_API + '/Authentication',
                              session_token)

    return client


def login_user(protocol, host, port, username, login=False):
    """ Login with the given user name.

    If login is False the user will be logged out.
    """
    session = UserCredentials()
    auth_client = ThriftAuthHelper(protocol, host, port,
                                   '/v' + CLIENT_API + '/Authentication')

    if not login:
        logout_done = auth_client.destroySession()
        if logout_done:
            session.save_token(host, port, None, True)
            LOG.info("Successfully logged out.")
        return

    try:
        handshake = auth_client.getAuthParameters()

        if not handshake.requiresAuthentication:
            LOG.info("This server does not require privileged access.")
            return

    except TApplicationException:
        LOG.info("This server does not support privileged access.")
        return

    methods = auth_client.getAcceptedAuthMethods()
    # Attempt username-password auth first.
    if 'Username:Password' in str(methods):

        # Try to use a previously saved credential from configuration file if
        # autologin is enabled.
        saved_auth = None
        if session.is_autologin_enabled():
            saved_auth = session.get_auth_string(host, port)

        if saved_auth:
            LOG.info("Logging in using preconfigured credentials...")
            username = saved_auth.split(":")[0]
            pwd = saved_auth.split(":")[1]
            check_preconfigured_username(username, host, port)
        else:
            LOG.info("Logging in using credentials from command line...")
            pwd = getpass.getpass(
                "Please provide password for user '{}': ".format(username))

        LOG.debug("Trying to login as %s to %s:%d", username, host, port)
        try:
            session_token = auth_client.performLogin("Username:Password",
                                                     username + ":" +
                                                     pwd)

            session.save_token(host, port, session_token)
            LOG.info("Server reported successful authentication.")
        except codechecker_api_shared.ttypes.RequestFailed as reqfail:
            LOG.error("Authentication failed! Please check your credentials.")
            LOG.error(reqfail.message)
            sys.exit(1)
    else:
        LOG.critical("No authentication methods were reported by the server "
                     "that this client could support.")
        sys.exit(1)


def perform_auth_for_handler(auth_client, host, port, manager):
    # Before actually communicating with the server,
    # we need to check authentication first.

    try:
        auth_response = auth_client.getAuthParameters()
    except TApplicationException:
        auth_response = AuthTypes.HandshakeInformation()
        auth_response.requiresAuthentication = False

    if auth_response.requiresAuthentication and \
            not auth_response.sessionStillActive:

        if manager.is_autologin_enabled():
            auto_auth_string = manager.get_auth_string(host, port)
            if auto_auth_string:
                LOG.info("Logging in using pre-configured credentials...")

                username = auto_auth_string.split(':')[0]
                check_preconfigured_username(username, host, port)

                LOG.debug("Trying to login as '%s' to '%s:%d'", username,
                          host, port)

                # Try to automatically log in with a saved credential
                # if it exists for the server.
                try:
                    session_token = auth_client.performLogin(
                        "Username:Password",
                        auto_auth_string)
                    manager.save_token(host, port, session_token)
                    LOG.info("Authentication successful.")
                    return session_token
                except codechecker_api_shared.ttypes.RequestFailed:
                    pass

        if manager.is_autologin_enabled():
            LOG.error("Invalid pre-configured credentials.")
            LOG.error("Your password has been changed or personal access "
                      "token has been removed which is used by your "
                      "\"%s\" file. Please "
                      "remove or change invalid credentials.",
                      env.get_password_file())
        else:
            LOG.error("Access denied. This server requires "
                      "authentication.")
            LOG.error("Please log in onto the server using 'CodeChecker "
                      "cmd login'.")
        sys.exit(1)


def setup_product_client(protocol, host, port, auth_client=None,
                         product_name=None,
                         session_token=None):
    """Setup the Thrift client for the product management endpoint."""
    cred_manager = UserCredentials()
    session_token = cred_manager.get_token(host, port)

    if not session_token:
        auth_client = setup_auth_client(protocol, host, port)
        session_token = perform_auth_for_handler(auth_client, host, port,
                                                 cred_manager)

    if not product_name:
        # Attach to the server-wide product service.
        product_client = ThriftProductHelper(
            protocol, host, port,
            '/v' + CLIENT_API + '/Products',
            session_token,
            lambda: get_new_token(protocol, host, port, cred_manager))
    else:
        # Attach to the product service and provide a product name
        # as "viewpoint" from which the product service is called.
        product_client = ThriftProductHelper(
            protocol, host, port,
            '/' + product_name + '/v' + CLIENT_API + '/Products',
            session_token,
            lambda: get_new_token(protocol, host, port, cred_manager))

        # However, in this case, the specified product might not exist,
        # which means we can't communicate with the server orderly.
        if not product_client.getPackageVersion() or \
                not product_client.getCurrentProduct():
            LOG.error("The product '%s' cannot be communicated with. It "
                      "either doesn't exist, or the server's configuration "
                      "is bogus.", product_name)
            sys.exit(1)

    return product_client


def get_new_token(protocol, host, port, cred_manager):
    """ Get a new session token from the remote server. """
    auth_client = setup_auth_client(protocol, host, port)
    return perform_auth_for_handler(auth_client, host, port, cred_manager)


def setup_client(product_url) -> ThriftResultsHelper:
    """Setup the Thrift Product or Service client and
    check API version and authentication needs.
    """

    try:
        protocol, host, port, product_name = split_product_url(product_url)
    except ValueError:
        LOG.error("Malformed product URL was provided. A valid product URL "
                  "looks like this: 'http://my.server.com:80/ProductName'.")
        sys.exit(2)  # 2 for argument error.

    # Check if local token is available.
    cred_manager = UserCredentials()
    session_token = cred_manager.get_token(host, port)

    # Local token is missing ask remote server.
    if not session_token:
        session_token = get_new_token(protocol, host, port, cred_manager)

    LOG.debug("Initializing client connecting to %s:%d/%s done.",
              host, port, product_name)

    return ThriftResultsHelper(
        protocol, host, port,
        '/' + product_name + '/v' + CLIENT_API + '/CodeCheckerService',
        session_token,
        lambda: get_new_token(protocol, host, port, cred_manager))



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Parse the plist output of an analyzer
"""

import importlib
import logging
import os
import plistlib
import traceback
import sys

from plistlib import _PlistParser  # type: ignore
from typing import Any, BinaryIO, Dict, List, Optional, Tuple

from xml.parsers.expat import ExpatError

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from mypy_extensions import TypedDict

from codechecker_report_converter.report import \
    BugPathEvent, BugPathPosition, \
    File, \
    MacroExpansion, \
    Range, Report, \
    UnknownChecker, \
    get_or_create_file
from codechecker_report_converter.report.hash import get_report_hash, HashType
from codechecker_report_converter.report.parser.base import AnalyzerInfo, \
    BaseParser, get_tool_info


LOG = logging.getLogger('report-converter')


EXTENSION = 'plist'

PlistItem = Any


class _LXMLPlistEventHandler:
    """
    Basic lxml event handler.
    """
    def start(self, tag, attrib):
        pass

    def end(self, tag):
        pass

    def data(self, data):
        pass

    def comment(self, text):
        pass

    def close(self):
        return "closed!"


class _LXMLPlistParser(_PlistParser):
    """
    Plist parser which uses the lxml library to parse XML data.

    The benefit of this library that this is faster than other libraries so it
    will improve the performance of the plist parsing.
    """
    def __init__(self, dict_type=dict):
        # Since Python 3.9 plistlib._PlistParser.__init__ has changed:
        # https://github.com/python/cpython/commit/ce81a925ef
        # To be backward compatible with old interpreters we need to call this
        # function based on conditions:
        params = _PlistParser.__init__.__code__.co_varnames
        # Before 3.9 interpreter. When a newer interpreter is used, pylint will
        # complain that too many arguments are used to invoke __init__, but
        # with newer interpreters, this is deadcode.
        if len(params) == 3 and "use_builtin_types" in params:
            # Before 3.9 interpreter.
            # pylint: disable=E1121
            _PlistParser.__init__(self, True, dict_type)
        else:
            # After 3.9 interpreter.
            _PlistParser.__init__(self, dict_type)  # pylint: disable=E1120

        self.event_handler = _LXMLPlistEventHandler()
        self.event_handler.start = self.handle_begin_element
        self.event_handler.end = self.handle_end_element
        self.event_handler.data = self.handle_data

        from lxml.etree import XMLParser  # pylint: disable=no-name-in-module
        self.parser = XMLParser(target=self.event_handler)

    def parse(self, fileobj):
        # pylint: disable=no-name-in-module
        from lxml.etree import parse, XMLSyntaxError

        try:
            parse(fileobj, self.parser)
        except XMLSyntaxError as ex:
            LOG.error("Invalid plist file '%s': %s", fileobj.name, ex)
            return

        return self.root


class DiagLoc(TypedDict):
    line: int
    col: int


class DiagEdge(TypedDict):
    start: Tuple[DiagLoc, DiagLoc]
    end: Tuple[DiagLoc, DiagLoc]


class DiagPath(TypedDict):
    kind: str
    message: str
    location: DiagLoc
    edges: List[DiagEdge]


def is_same_control_item(
    curr: DiagPath,
    prev: DiagPath
) -> bool:
    """ True if the given diag paths are same. """
    curr_start_range_begin = curr['edges'][0]['start'][0]
    curr_start_range_end = curr['edges'][0]['start'][1]

    prev_end_range_begin = prev['edges'][0]['end'][0]
    prev_end_range_end = prev['edges'][0]['end'][1]

    return curr_start_range_begin == prev_end_range_begin and \
        curr_start_range_end == prev_end_range_end


def parse(fp: BinaryIO):
    """
    Read a .plist file. Return the unpacked root object (which usually is a
    dictionary).

    Use 'lxml' library to read the given plist file if it is available,
    otherwise use 'plistlib' library.
    """
    try:
        importlib.import_module('lxml')
        parser = _LXMLPlistParser()
        return parser.parse(fp)
    except (ExpatError, TypeError, AttributeError) as err:
        LOG.warning('Invalid plist file')
        LOG.warning(err)
        return
    except ImportError:
        LOG.debug("lxml library is not available. Use plistlib to parse plist "
                  "files.")

    try:
        return plistlib.load(fp)
    except (ExpatError, TypeError, AttributeError, ValueError,
            plistlib.InvalidFileException) as err:
        LOG.warning('Invalid plist file')
        LOG.warning(err)
        return


def get_file_index_map(
    plist: Any,
    source_dir_path: str,
    file_cache: Dict[str, File]
) -> Dict[int, File]:
    """ Get file index map from the given plist object. """
    file_index_map: Dict[int, File] = {}

    for i, orig_file_path in enumerate(plist.get('files', [])):
        file_path = os.path.normpath(os.path.join(
            source_dir_path, orig_file_path))
        file_index_map[i] = get_or_create_file(file_path, file_cache)

    return file_index_map


class Parser(BaseParser):
    def get_reports(
        self,
        analyzer_result_file_path: str,
        source_dir_path: Optional[str] = None
    ) -> List[Report]:
        """ Get reports from the given analyzer result file. """
        reports: List[Report] = []

        if not source_dir_path:
            source_dir_path = os.path.dirname(analyzer_result_file_path)

        try:
            with open(analyzer_result_file_path, 'rb') as fp:
                plist = parse(fp)

            if not plist:
                return reports

            metadata = plist.get('metadata')

            files = get_file_index_map(
                plist, source_dir_path, self._file_cache)

            for diag in plist.get('diagnostics', []):
                report = self.__create_report(
                    analyzer_result_file_path, diag, files, metadata)

                if report.report_hash is None:
                    report.report_hash = get_report_hash(
                        report, HashType.PATH_SENSITIVE)

                reports.append(report)
        except KeyError as ex:
            LOG.warning("Failed to get file path id! Found files: %s. "
                        "KeyError: %s", files, ex)
        except IndexError as iex:
            LOG.warning("Indexing error during processing plist file %s",
                        analyzer_result_file_path)
            LOG.warning(type(iex))
            LOG.warning(repr(iex))
            _, _, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        except Exception as ex:
            LOG.warning("Error during processing reports from the plist "
                        "file: %s", analyzer_result_file_path)
            traceback.print_exc()
            LOG.warning(type(ex))
            LOG.warning(ex)
        finally:
            return reports

    def __create_report(
        self,
        analyzer_result_file_path: str,
        diag: Dict,
        files: Dict[int, File],
        metadata: Dict[str, Any]
    ) -> Report:
        location = diag.get("location", {})
        checker_name = diag.get("check_name", UnknownChecker[1])
        analyzer_name = self.__get_analyzer_name(checker_name, metadata)
        severity = self.get_severity(checker_name)

        report_annotation = diag["report-annotation"] \
            if "report-annotation" in diag else None

        return Report(
            analyzer_result_file_path=analyzer_result_file_path,
            file=files[location['file']],
            line=location.get('line', -1),
            column=location.get('col', -1),
            message=diag.get('description', ''),
            checker_name=checker_name,
            severity=severity,
            report_hash=diag.get('issue_hash_content_of_line_in_context'),
            analyzer_name=analyzer_name,
            category=diag.get('category'),
            source_line=None,
            bug_path_events=self.__get_bug_path_events(diag, files),
            bug_path_positions=self.__get_bug_path_positions(diag, files),
            notes=self.__get_notes(diag, files),
            macro_expansions=self.__get_macro_expansions(diag, files),
            annotations=report_annotation)

    def __get_analyzer_name(
        self,
        checker_name: str,
        metadata: Dict[str, Any]
    ) -> Optional[str]:
        """ Get analyzer name for the given checker name. """
        if metadata:
            name = metadata.get("analyzer", {}).get("name")
            if name:
                return name

        if checker_name.startswith('clang-diagnostic-'):
            return 'clang-tidy'

        return None

    def __get_bug_event_locations(self, item: PlistItem):
        """ Get bug path position for the given plist item. """
        location = item['location']
        ranges = item.get("ranges")

        # Range can provide more precise location information.
        # Use that if available.
        if ranges:
            return location, ranges[0][0], ranges[0][1]

        return location, location, location

    def __get_bug_path_events(
        self,
        diag,
        files: Dict[int, File]
    ) -> List[BugPathEvent]:
        """ Get bug path events. """
        events = []

        for item in diag.get('path', []):
            if item.get('kind') != 'event':
                continue

            location, start_loc, end_loc = self.__get_bug_event_locations(item)
            events.append(BugPathEvent(
                message=item['message'],
                file=files[location['file']],
                line=location['line'],
                column=location['col'],
                range=Range(
                    start_loc['line'], start_loc['col'],
                    end_loc['line'], end_loc['col'])))

        return events

    def __get_bug_path_positions(
        self,
        diag,
        files: Dict[int, File]
    ) -> List[BugPathPosition]:
        """ Get bug path positions.

        In plist file the source and target of the arrows are provided as
        starting and ending ranges of the arrow. The path A->B->C is given as
        A->B and B->C, thus range B is provided twice if multiple control event
        kinds are followed each other. So in the loop we will not store the
        start point if the previous path event was a control event.
        """
        bug_path_positions = []

        prev_control_item = None
        for item in diag.get('path', []):
            if item.get('kind') != 'control':
                continue

            try:
                edges = item['edges'][0]

                edge = None
                if prev_control_item:
                    if not is_same_control_item(item, prev_control_item):
                        edge = edges['start']
                else:
                    edge = edges['start']

                if edge:
                    bug_path_positions.append(BugPathPosition(
                        file=files[edge[1]['file']],
                        range=Range(
                            edge[0]['line'], edge[0]['col'],
                            edge[1]['line'], edge[1]['col'])))

                bug_path_positions.append(BugPathPosition(
                    file=files[edges['end'][1]['file']],
                    range=Range(
                        edges['end'][0]['line'], edges['end'][0]['col'],
                        edges['end'][1]['line'], edges['end'][1]['col'])))

                prev_control_item = item
            except IndexError:
                # Edges might be empty nothing can be stored.
                continue

        return bug_path_positions

    def __get_notes(
        self,
        diag,
        files: Dict[int, File]
    ) -> List[BugPathEvent]:
        """ Get notes. """
        notes = []

        for note in diag.get('notes', []):
            if not note['message']:
                continue

            location, start_loc, end_loc = self.__get_bug_event_locations(note)
            notes.append(BugPathEvent(
                message=note['message'],
                file=files[location['file']],
                line=location['line'],
                column=location['col'],
                range=Range(
                    start_loc['line'], start_loc['col'],
                    end_loc['line'], end_loc['col'])))

        return notes

    def __get_macro_expansions(
        self,
        diag,
        files: Dict[int, File]
    ) -> List[MacroExpansion]:
        """ Get macro expansion. """
        macro_expansions = []

        for macro in diag.get('macro_expansions', []):
            if not macro['expansion']:
                continue

            location, start_loc, end_loc = self.__get_bug_event_locations(
                macro)
            macro_expansions.append(MacroExpansion(
                message=macro['expansion'],
                name=macro['name'],
                file=files[location['file']],
                line=location['line'],
                column=location['col'],
                range=Range(
                    start_loc['line'], start_loc['col'],
                    end_loc['line'], end_loc['col'])))

        return macro_expansions

    def convert(
        self,
        reports: List[Report],
        analyzer_info: Optional[AnalyzerInfo] = None
    ):
        """ Converts the given reports. """
        tool_name, tool_version = get_tool_info()

        data: Dict[str, Any] = {
            'files': [],
            'diagnostics': [],
            'metadata': {
                'generated_by': {
                    'name': tool_name,
                    'version': tool_version
                }
            }
        }

        if analyzer_info:
            data['metadata']['analyzer'] = {'name': analyzer_info.name}

        files = set()
        for report in reports:
            files.update(report.original_files)

        file_index_map: Dict[str, int] = {}
        for idx, file_path in enumerate(sorted(files)):
            data['files'].append(file_path)
            file_index_map[file_path] = idx

        for report in reports:
            diagnostic = {
                'location': self._create_location(
                    report.line, report.column,
                    file_index_map[report.file.original_path]),
                'issue_hash_content_of_line_in_context': report.report_hash,
                'check_name': report.checker_name,
                'description': report.message,
                'category': report.category or 'unknown'
            }

            if report.analyzer_name:
                diagnostic['type'] = report.analyzer_name

            path = []
            if report.bug_path_positions:
                for i in range(len(report.bug_path_positions) - 1):
                    start = report.bug_path_positions[i]
                    end = report.bug_path_positions[i + 1]
                    if start.range and end.range:
                        edge = self._create_control_edge(
                            start.range, start.file,
                            end.range, end.file,
                            file_index_map)
                        path.append(self._create_control_edges([edge]))
            elif len(report.bug_path_events) > 1:
                # Create bug path positions from bug path events.
                for i in range(len(report.bug_path_events) - 1):
                    start = report.bug_path_events[i]
                    start_range = self._get_bug_path_event_range(start)

                    end = report.bug_path_events[i + 1]
                    end_range = self._get_bug_path_event_range(end)

                    if start_range == end_range:
                        continue

                    edge = self._create_control_edge(
                        start_range, start.file,
                        end_range, end.file,
                        file_index_map)
                    path.append(self._create_control_edges([edge]))

            # Add bug path events after control points.
            if report.bug_path_events:
                for event in report.bug_path_events:
                    path.append(self._create_event(event, file_index_map))

            diagnostic['path'] = path

            if report.notes:
                diagnostic['notes'] = []
                for note in report.notes:
                    diagnostic['notes'].append(
                        self._create_note(note, file_index_map))

            if report.macro_expansions:
                diagnostic['macro_expansions'] = []
                for macro_expansion in report.macro_expansions:
                    diagnostic['macro_expansions'].append(
                        self._create_macro_expansion(
                            macro_expansion, file_index_map))

            if report.annotations:
                diagnostic["report-annotation"] = dict()
                for key, value in report.annotations.items():
                    diagnostic["report-annotation"][key] = value

            data['diagnostics'].append(diagnostic)

        return data

    def write(self, data: Any, output_file_path: str):
        """ Creates an analyzer output file from the given data. """
        try:
            with open(output_file_path, 'wb') as f:
                plistlib.dump(data, f)
        except TypeError as err:
            LOG.error('Failed to write plist file: %s', output_file_path)
            LOG.error(err)
            import traceback
            traceback.print_exc()

    def _get_bug_path_event_range(self, event: BugPathEvent) -> Range:
        """ Get range for bug path event. """
        if event.range:
            return event.range

        return Range(event.line, event.column, event.line, event.column)

    def _create_location(
        self,
        line: int,
        column: int,
        file_index: int
    ):
        """ Create a location section from the message. """
        return {'line': line, 'col': column, 'file': file_index}

    def _create_event(
        self,
        event: BugPathEvent,
        file_index_map: Dict[str, int]
    ):
        """ Create an event. """
        data = {
            'kind': 'event',
            'location': self._create_location(
                event.line, event.column,
                file_index_map[event.file.original_path]),
            'depth': 0,
            'message': event.message}

        if event.range:
            data['ranges'] = [self._create_range(
                event.range, file_index_map[event.file.original_path])]

        return data

    def _create_control_edges(self, edges: List[Dict]) -> Dict:
        """ """
        return {'kind': 'control', 'edges': edges}

    def _create_control_edge(
        self,
        start_range: Range,
        start_file: File,
        end_range: Range,
        end_file: File,
        file_index_map: Dict[str, int]
    ) -> Dict:
        """ Creates a control point. """
        return {
            'start': self._create_range(
                start_range, file_index_map[start_file.original_path]),
            'end': self._create_range(
                end_range, file_index_map[end_file.original_path])}

    def _create_note(
        self,
        note: BugPathEvent,
        file_index_map: Dict[str, int]
    ):
        """ Creates a note. """
        data = {
            'location': self._create_location(
                note.line, note.column,
                file_index_map[note.file.original_path]),
            'message': note.message}

        if note.range:
            data['ranges'] = [self._create_range(
                note.range, file_index_map[note.file.original_path])]

        return data

    def _create_range(
        self,
        range: Range,
        file_idx: int
    ) -> List:
        """ Creates a range. """
        return [
            self._create_location(range.start_line, range.start_col, file_idx),
            self._create_location(range.end_line, range.end_col, file_idx)]

    def _create_macro_expansion(
        self,
        macro_expansion: MacroExpansion,
        file_index_map: Dict[str, int]
    ):
        """ Creates a macro expansion. """
        return {
            'name': macro_expansion.name,
            'expansion': macro_expansion.message,
            'location': self._create_location(
                macro_expansion.line, macro_expansion.column,
                file_index_map[macro_expansion.file.original_path])}

    def replace_report_hash(
        self,
        plist_file_path: str,
        hash_type=HashType.CONTEXT_FREE
    ):
        """
        Override hash in the given file by using the given version hash.
        """
        try:
            with open(plist_file_path, 'rb+') as f:
                plist = plistlib.load(f)
                f.seek(0)
                f.truncate()

                metadata = plist.get('metadata')
                analyzer_result_dir_path = os.path.dirname(plist_file_path)

                file_cache: Dict[str, File] = {}
                files = get_file_index_map(
                    plist, analyzer_result_dir_path, file_cache)

                for diag in plist['diagnostics']:
                    report = self.__create_report(
                        plist_file_path, diag, files, metadata)
                    diag['issue_hash_content_of_line_in_context'] = \
                        get_report_hash(report, hash_type)

                plistlib.dump(plist, f)
        except (TypeError, AttributeError,
                plistlib.InvalidFileException) as err:
            LOG.warning('Failed to process plist file: %s wrong file format?',
                        plist_file_path)
            LOG.warning(err)
        except IndexError as iex:
            LOG.warning('Indexing error during processing plist file %s',
                        plist_file_path)
            LOG.warning(type(iex))
            LOG.warning(repr(iex))
            _, _, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback, limit=1, file=sys.stdout)
        except Exception as ex:
            LOG.warning('Error during processing reports from the plist '
                        'file: %s', plist_file_path)
            traceback.print_exc()
            LOG.warning(type(ex))
            LOG.warning(ex)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import os
import xml.etree.ElementTree as ET

from typing import Dict, List, Optional

from codechecker_report_converter.report import File, get_or_create_file, \
    Report

from ..analyzer_result import AnalyzerResultBase


LOG = logging.getLogger('report-converter')


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of Roslynator. """

    TOOL_NAME = 'roslynator'
    NAME = 'Roslynator.DotNet.Cli'
    URL = 'https://github.com/JosefPihrt/Roslynator' \
        + '#roslynator-command-line-tool-'

    def __init__(self):
        super(AnalyzerResult, self).__init__()
        self.__file_cache: Dict[str, File] = {}

    def get_reports(
        self,
        file_path: str
    ) -> List[Report]:
        """ Get reports from the given analyzer result. """
        reports: List[Report] = []

        root = self.__parse_analyzer_result(file_path)
        if root is None:
            return reports

        for diag in root.findall('./CodeAnalysis/Projects/*/Diagnostics/*'):
            report = self.__parse_diag(diag, file_path)
            if report is not None:
                reports.append(report)

        return reports

    def __parse_analyzer_result(
        self,
        analyzer_result: str
    ) -> Optional[ET.Element]:
        """ Parse the given analyzer result xml file.

        Returns the root element of the parsed tree or None if something goes
        wrong.
        """
        try:
            tree = ET.parse(analyzer_result)
            return tree.getroot()
        except OSError:
            LOG.error("Analyzer result does not exist: %s", analyzer_result)
            return None
        except ET.ParseError:
            LOG.error("Failed to parse the given analyzer result '%s'. Please "
                      "give a valid xml file with messages generated by "
                      "Roslynator.", analyzer_result)
            return None

    def __parse_diag(
        self,
        diag,
        input_file_path: str
    ) -> Optional[Report]:
        """ Parse the given Roslynator diagnostic

        Returns the Report from the parsed diagnostic or None if something goes
        wrong.
        """
        id = diag.attrib.get('Id')

        filePathElement = diag.find('FilePath')
        if filePathElement is None:
            LOG.warning("Diagnostic does not belong to a file: %s", id)
            return None

        source_file_path = os.path.join(os.path.dirname(input_file_path),
                                        filePathElement.text)
        if not os.path.exists(source_file_path):
            LOG.warning("Source file does not exist: %s", source_file_path)
            return None

        message = diag.find('Message').text
        location = diag.find('Location')
        line = location.attrib.get('Line')
        column = location.attrib.get('Character')

        return Report(
            get_or_create_file(source_file_path, self.__file_cache),
            int(line),
            int(column),
            message,
            id
        )



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Util module.
"""
import itertools
import json
import os
from typing import TextIO

import portalocker

from codechecker_common.logger import get_logger

LOG = get_logger('system')


def arg_match(options, args):
    """Checks and selects the option string specified in 'options'
    that are present in parameter 'args'."""
    matched_args = []
    for option in options:
        if any([arg if option.startswith(arg) else None
                for arg in args]):
            matched_args.append(option)
            continue

    return matched_args


def clamp(min_: int, value: int, max_: int) -> int:
    """Clamps ``value`` such that ``min_ <= value <= max_``."""
    if min_ > max_:
        raise ValueError("min <= max required")
    return min(max(min_, value), max_)


def chunks(iterator, n):
    """
    Yield the next chunk if an iterable object. A chunk consists of maximum n
    elements.
    """
    iterator = iter(iterator)
    for first in iterator:
        rest_of_chunk = itertools.islice(iterator, 0, n - 1)
        yield itertools.chain([first], rest_of_chunk)


def load_json(path: str, default=None, lock=False, display_warning=True):
    """
    Load the contents of the given file as a JSON and return it's value,
    or default if the file can't be loaded.

    path -- JSON file path to load.
    defaut -- Value to return if JSON can't be loaded for some reason (e.g
              file doesn't exist, bad JSON format, etc.)
    lock -- Use portalocker to lock the JSON file for exclusive use.
    display_warning -- Display warning messages why the JSON file can't be
                       loaded (e.g. bad format, failed to open file, etc.)
    """

    ret = default
    try:
        with open(path, 'r', encoding='utf-8', errors='ignore') as handle:
            if lock:
                portalocker.lock(handle, portalocker.LOCK_SH)

            ret = json.load(handle)

            if lock:
                portalocker.unlock(handle)
    except IOError as ex:
        if display_warning:
            LOG.warning("Failed to open json file: %s", path)
            LOG.warning(ex)
    except OSError as ex:
        if display_warning:
            LOG.warning("Failed to open json file: %s", path)
            LOG.warning(ex)
    except ValueError as ex:
        if display_warning:
            LOG.warning("%s is not a valid json file.", path)
            LOG.warning(ex)
    except TypeError as ex:
        if display_warning:
            LOG.warning('Failed to process json file: %s', path)
            LOG.warning(ex)

    return ret


def get_linef(fp: TextIO, line_no: int) -> str:
    """'fp' should be (readable) file object.
    Return the line content at line_no or an empty line
    if there is less lines than line_no.
    """
    fp.seek(0)
    for line in fp:
        line_no -= 1
        if line_no == 0:
            return line
    return ''


def path_for_fake_root(full_path: str, root_path: str = '/') -> str:
    """Normalize and sanitize full_path, then make it relative to root_path."""
    relative_path = os.path.relpath(full_path, '/')
    fake_root_path = os.path.join(root_path, relative_path)
    return os.path.realpath(fake_root_path)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Clang Static Analyzer related functions.
"""

import os
import subprocess

from codechecker_common.logger import get_logger
from codechecker_analyzer import host_check
from codechecker_analyzer.analyzers.clangsa import version

LOG = get_logger('analyzer.clangsa')

CTU_ON_DEMAND_OPTION_NAME = 'ctu-invocation-list'


def ctu_mapping(clang_version_info):
    """Clang version dependent ctu mapping tool path and mapping file name.

    The path of the mapping tool, which is assumed to be located
    inside the installed directory of the analyzer. Certain binary
    distributions can postfix the tool name with the major version
    number, the number and the tool name being separated by a dash. By
    default the shorter name is looked up, then if it is not found the
    postfixed.
    """
    if not clang_version_info:
        LOG.debug("No clang version information."
                  "Can not detect ctu mapping tool.")
        return None, None

    old_mapping_tool_name = 'clang-func-mapping'
    old_mapping_file_name = 'externalFnMap.txt'

    new_mapping_tool_name = 'clang-extdef-mapping'
    new_mapping_file_name = 'externalDefMap.txt'

    major_version = clang_version_info.major_version

    if major_version > 7:
        tool_name = new_mapping_tool_name
        mapping_file = new_mapping_file_name
    else:
        tool_name = old_mapping_tool_name
        mapping_file = old_mapping_file_name

    installed_dir = clang_version_info.installed_dir

    tool_path = os.path.join(installed_dir, tool_name)

    if os.path.isfile(tool_path):
        return tool_path, mapping_file

    LOG.debug(
        "Mapping tool '%s' suggested by autodetection is not found in "
        "directory reported by Clang '%s'. Trying with version-postfixed "
        "filename...", tool_path, installed_dir)

    postfixed_tool_path = ''.join([tool_path, '-', str(major_version)])

    if os.path.isfile(postfixed_tool_path):
        return postfixed_tool_path, mapping_file

    LOG.debug(
        "Postfixed mapping tool '%s' suggested by autodetection is not "
        "found in directory reported by Clang '%s'.",
        postfixed_tool_path, installed_dir)
    return None, None


def invoke_binary_checked(binary_path, args=None, environ=None):
    """
    Invoke the binary with the specified args, and return the output if the
    command finished running with zero exit code. Return False otherwise.
    Possible usage can be used to check the existence binaries.

    :param binary_path: The path to the executable to invoke
    :param args: The arguments of the invocation
    :type binary_path: str
    :type args: list
    :rtype str
    """

    args = args or []
    invocation = [binary_path]
    invocation.extend(args)
    try:
        output = subprocess.check_output(
            invocation,
            env=environ,
            encoding="utf-8",
            errors="ignore")
    except (subprocess.CalledProcessError, OSError) as e:
        LOG.debug('Command invocation failed because of non-zero exit code!'
                  'Details: %s', str(e))
        return False
    return output


class CTUAutodetection:
    """
    CTUAutodetection is responsible for providing the availability information
    of CTU feature, the the relevant mapping tool path and the mapping file
    name.
    """

    def __init__(self, analyzer_binary, environ):
        self.__analyzer_binary = analyzer_binary
        self.environ = environ
        self.__analyzer_version_info = None

        if self.__analyzer_binary is None:
            LOG.debug(
                'Trying to detect CTU capability, but analyzer binary is not '
                'set!')
            return None

        analyzer_version = invoke_binary_checked(
            self.__analyzer_binary, ['--version'], self.environ)

        if analyzer_version is False:
            LOG.debug('Failed to invoke command to get Clang version!')
            return None

        version_parser = version.ClangVersionInfoParser(self.__analyzer_binary)
        version_info = version_parser.parse(analyzer_version)

        if not version_info:
            LOG.debug('Failed to parse Clang version information!')
            return None

        self.__analyzer_version_info = version_info

    @property
    def analyzer_version_info(self):
        """
        Returns the relevant parameters of the analyzer by parsing the
        output of the analyzer binary when called with version flag.
        """
        if not self.__analyzer_version_info:
            return False

        return self.__analyzer_version_info

    @property
    def major_version(self):
        """
        Returns the major version of the analyzer, which is used for
        CTU analysis.
        """
        return self.analyzer_version_info.major_version

    @property
    def installed_dir(self):
        """
        Returns the installed directory of the analyzer, which is used for
        CTU analysis.
        """
        return self.analyzer_version_info.installed_dir

    @property
    def mapping_tool_path(self):
        """Return the path to the mapping tool."""
        tool_path, _ = ctu_mapping(self.analyzer_version_info)

        if tool_path:
            return tool_path
        return False

    @property
    def display_progress(self):
        """
        Return analyzer args if it is capable to display ctu progress.

        Returns None if the analyzer can not display ctu progress.
        The ctu display progress arguments depend on
        the clang analyzer version.
        """

        if not self.analyzer_version_info:
            return None
        ctu_display_progress_args = ['-Xclang',
                                     '-analyzer-config',
                                     '-Xclang',
                                     'display-ctu-progress=true']

        ok = host_check.has_analyzer_config_option(
            self.__analyzer_binary, "display-ctu-progress")
        if not ok:
            return None
        return ctu_display_progress_args

    @property
    def mapping_file_name(self):
        """
        Returns the installed directory of the analyzer, which is used for
        CTU analysis.
        """

        _, mapping_file_name = ctu_mapping(self.analyzer_version_info)

        if mapping_file_name:
            return mapping_file_name
        return False

    @property
    def is_ctu_capable(self):
        """
        Detects if the current clang is CTU compatible. Tries to autodetect
        the correct one based on clang version.
        """

        tool_path = self.mapping_tool_path

        if not tool_path:
            return False

        return invoke_binary_checked(tool_path, ['-version'], self.environ) \
            is not False

    @property
    def is_on_demand_ctu_available(self):
        """
        Detects if the current Clang supports on-demand parsing of ASTs for
        CTU analysis.
        """

        analyzer_options = invoke_binary_checked(
            self.__analyzer_binary, ['-cc1', '-analyzer-config-help'],
            self.environ)

        if analyzer_options is False:
            return False

        return CTU_ON_DEMAND_OPTION_NAME in analyzer_options



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
""" Helper and converter functions for plain text format. """

import logging
import math
import os
import sys

from collections import defaultdict
from typing import Dict, List, Optional, Set

from codechecker_report_converter.report import BugPathEvent, \
    InvalidFileContentMsg, MacroExpansion, Report


LOG = logging.getLogger('report-converter')


def __get_source_file_for_analyzer_result_file(
    analyzer_result_file_path: str,
    metadata: Optional[Dict]
) -> Optional[str]:
    """ Get source file for the given analyzer result file. """
    if not metadata:
        return None

    result_source_files = {}
    if 'result_source_files' in metadata:
        result_source_files = metadata['result_source_files']
    else:
        for tool in metadata.get('tools', {}):
            result_src_files = tool.get('result_source_files', {})
            result_source_files.update(result_src_files.items())

    if analyzer_result_file_path in result_source_files:
        return result_source_files[analyzer_result_file_path]

    return None


def format_main_report(report: Report) -> str:
    """ Format bug path event. """
    line = report.source_line
    if line == '':
        return ''

    marker_line = line[0:(report.column - 1)]
    marker_line = ' ' * (len(marker_line) + marker_line.count('\t'))

    line = line.replace('\t', '  ')

    return f"{line}{marker_line}^"


def format_report(report: Report, content_is_not_changed: bool) -> str:
    """ Format main report. """
    file_path = report.bug_path_events[-1].file.path
    out = f"[{report.severity}] {file_path}:{report.line}:{report.column}: " \
          f"{report.message} [{report.checker_name}]"

    if content_is_not_changed and report.review_status and \
            report.review_status.status != 'unreviewed':
        out += f" [{report.review_status.formatted_status()}]"

    return out


def format_note(note: BugPathEvent) -> str:
    """ Format bug path note. """
    file_name = os.path.basename(note.file.path)
    return f"{file_name}:{note.line}:{note.column}: {note.message}"


def format_macro_expansion(macro: MacroExpansion) -> str:
    """ Format macro expansions. """
    file_name = os.path.basename(macro.file.path)
    return f"{file_name}:{macro.line}:{macro.column}: Macro '{macro.name}' " \
           f"expanded to '{macro.message}'"


def format_event(event: BugPathEvent) -> str:
    """ Format bug path event. """
    file_name = os.path.basename(event.file.path)
    return f"{file_name}:{event.line}:{event.column}: {event.message}"


def get_index_format(lst: List) -> str:
    """ Get index format. """
    return f'    %{int(math.floor(math.log10(len(lst))) + 1)}d, '


def print_details(report: Report, output=sys.stdout):
    """ Print detail information from the given report. """
    output.write(f'  Report hash: {report.report_hash}\n')

    # Print out macros.
    if report.macro_expansions:
        output.write('  Macro expansions:\n')
        index_format = get_index_format(report.macro_expansions)
        for index, macro in enumerate(report.macro_expansions):
            output.write(index_format % (index + 1))
            output.write(f"{format_macro_expansion(macro)}\n")

    # Print out notes.
    if report.notes:
        output.write('  Notes:\n')
        index_format = get_index_format(report.notes)
        for index, note in enumerate(report.notes):
            output.write(index_format % (index + 1))
            output.write(f"{format_note(note)}\n")

    # Print out bug path events.
    output.write('  Steps:\n')
    index_format = get_index_format(report.bug_path_events)
    for index, event in enumerate(report.bug_path_events):
        output.write(index_format % (index + 1))
        output.write(f"{format_event(event)}\n")


def get_file_report_map(
    reports: List[Report],
    input_file_path: Optional[str] = None,
    metadata: Optional[Dict] = None
) -> Dict[str, List[Report]]:
    """ Get file report map. """
    file_report_map = defaultdict(list)
    for report in reports:
        file_report_map[report.file.path].append(report)

    if input_file_path:
        source_file = __get_source_file_for_analyzer_result_file(
            input_file_path, metadata)

        # Add source file to the map if it doesn't exists.
        if source_file:
            file_report_map[source_file]

    return file_report_map


def convert(
    review_status_handler,
    source_file_report_map: Dict[str, List[Report]],
    processed_file_paths: Optional[Set[str]] = None,
    print_steps: bool = False,
    output=sys.stdout
):
    """ Convert reports to plain text format. """
    if processed_file_paths is None:
        processed_file_paths = set()

    for file_path in sorted(source_file_report_map, key=str.casefold):
        num_of_active_reports = 0
        reports = sorted(source_file_report_map[file_path],
                         key=lambda report: report.line)
        for report in reports:
            # If file content is changed, do not print the source code comments
            # (if available) and instead of the source line, print a warning
            # message.
            content_is_not_changed = \
                report.file.original_path not in report.changed_files

            output.write(f"{format_report(report, content_is_not_changed)}\n")

            if content_is_not_changed:
                source_comment = review_status_handler.source_comment(report)
                if source_comment and source_comment.line:
                    output.write(f"{source_comment.line.rstrip()}\n")

                output.write(f"{format_main_report(report)}")
            else:
                output.write(InvalidFileContentMsg)

            output.write("\n")

            if print_steps:
                print_details(report)

            output.write("\n")
            output.flush()

            num_of_active_reports += 1

        file_name = os.path.basename(file_path)
        if num_of_active_reports:
            output.write(f"Found {num_of_active_reports} defect(s) in "
                         f"{file_name}\n\n")
        elif file_path not in processed_file_paths:
            output.write(f"Found no defects in {file_name}\n")

        processed_file_paths.add(file_path)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handle Thrift requests for server info.
"""

from codechecker_common.logger import get_logger
from codechecker_server.profiler import timeit

LOG = get_logger('server')


class ThriftServerInfoHandler:
    """
    Manages Thrift requests regarding server info.
    """

    def __init__(self, package_version):
        self.__package_version = package_version

    @timeit
    def getPackageVersion(self):
        """ Get package version. """
        return self.__package_version



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
SQLAlchemy ORM model for the product configuration database.
"""
from datetime import datetime
import sys

from sqlalchemy import Boolean, CHAR, Column, DateTime, Enum, ForeignKey, \
    Integer, MetaData, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.expression import false, true

from ..permissions import get_permissions

CC_META = MetaData(naming_convention={
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s"
})

# Create base class for ORM classes.
Base = declarative_base(metadata=CC_META)


class Product(Base):
    __tablename__ = 'products'

    id = Column(Integer, autoincrement=True, primary_key=True)
    endpoint = Column(String, nullable=False, unique=True)
    connection = Column(String, nullable=False)
    display_name = Column(String, nullable=False)
    description = Column(Text)
    run_limit = Column(Integer)
    report_limit = Column(Integer, nullable=False, server_default="500000")
    num_of_runs = Column(Integer, server_default="0")
    latest_storage_date = Column(DateTime, nullable=True)

    # Disable review status change on UI.
    is_review_status_change_disabled = Column(Boolean,
                                              server_default=false())

    confidentiality = Column(String, nullable=True)

    def __init__(self, endpoint, conn_str, name=None, description=None,
                 run_limit=None, report_limit=500000,
                 is_review_status_change_disabled=False,
                 confidentiality=None):
        self.endpoint = endpoint
        self.connection = conn_str
        self.display_name = name if name else endpoint
        self.description = description
        self.run_limit = run_limit
        self.report_limit = report_limit
        self.is_review_status_change_disabled = \
            True if is_review_status_change_disabled else False
        self.confidentiality = confidentiality


def __get_permission_names(scope=None):
    """
    Returns a list of strings which contains permission names.

    This function is used internally to set up the permission database.

    :param scope: One of the Permission class strings (e.g. 'SYSTEM'), which
      if given, filters the returned list of permissions to only definitions
      of the given scope.
    """

    return [perm.name for perm in get_permissions(scope)]


class SystemPermission(Base):
    __tablename__ = 'permissions_system'

    id = Column(Integer, autoincrement=True, primary_key=True)
    permission = Column(Enum(
        *sys.modules[__name__].__dict__['__get_permission_names']('SYSTEM'),
        name='sys_perms'))
    name = Column(String, nullable=False)
    is_group = Column(Boolean, nullable=False)

    def __init__(self, permission, name, is_group=False):
        self.permission = permission
        self.name = name
        self.is_group = is_group


class ProductPermission(Base):
    __tablename__ = 'permissions_product'

    id = Column(Integer, autoincrement=True, primary_key=True)
    permission = Column(Enum(
        *sys.modules[__name__].__dict__['__get_permission_names']('PRODUCT'),
        name='product_perms'))
    product_id = Column(Integer, ForeignKey('products.id',
                                            deferrable=False,
                                            initially='IMMEDIATE',
                                            ondelete='CASCADE'),
                        nullable=False)
    name = Column(String, nullable=False)
    is_group = Column(Boolean, nullable=False)

    def __init__(self, permission, product_id, name, is_group=False):
        self.permission = permission
        self.product_id = product_id
        self.name = name
        self.is_group = is_group


class Session(Base):
    __tablename__ = 'auth_sessions'

    id = Column(Integer, autoincrement=True, primary_key=True)

    user_name = Column(String)
    token = Column(CHAR(32), nullable=False, unique=True)

    # List of group names separated by semicolons.
    groups = Column(String)

    last_access = Column(DateTime, nullable=False)

    # Token description.
    description = Column(String)

    can_expire = Column(Boolean, server_default=true(), default=True)

    def __init__(self, token, user_name, groups, description=None,
                 can_expire=True):
        self.token = token
        self.user_name = user_name
        self.groups = groups
        self.description = description
        self.can_expire = can_expire
        self.last_access = datetime.now()


class Configuration(Base):
    __tablename__ = 'server_configurations'

    id = Column(Integer, autoincrement=True, primary_key=True)
    config_key = Column(String,
                        nullable=False)

    config_value = Column(String)

    def __init__(self, config_key, config_value):
        self.config_key = config_key
        self.config_value = config_value


IDENTIFIER = {
    'identifier': "ConfigDatabase",
    'orm_meta': CC_META
}



"""
Add check command to run history

Revision ID: 080349e895d7
Revises:     101a9cb747de
Create Date: 2018-03-12 15:10:24.652576
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '080349e895d7'
down_revision = '101a9cb747de'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('run_histories',
                  sa.Column('check_command', sa.Binary(), nullable=True))


def downgrade():
    op.drop_column('run_histories', 'check_command')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""Argument parsing related helper classes and functions."""


import argparse
import collections
import os
import re


AnalyzerConfig = collections.namedtuple(
    'AnalyzerConfig', ["analyzer", "option", "value"])
CheckerConfig = collections.namedtuple(
    "CheckerConfig", ["analyzer", "checker", "option", "value"])
AnalyzerBinary = collections.namedtuple(
    "AnalyzerBinary", ["analyzer", "path"])


class OrderedCheckersAction(argparse.Action):
    """
    Action to store enabled and disabled checkers
    and keep ordering from command line.

    Create separate lists based on the checker names for
    each analyzer.
    """

    # Users can supply invocation to 'codechecker-analyze' as follows:
    # -e core -d core.uninitialized -e core.uninitialized.Assign
    # We must support having multiple '-e' and '-d' options and the order
    # specified must be kept when the list of checkers are assembled for Clang.

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(OrderedCheckersAction, self).__init__(option_strings, dest,
                                                    **kwargs)

    def __call__(self, parser, namespace, value, option_string=None):

        if 'ordered_checkers' not in namespace:
            namespace.ordered_checkers = []
        ordered_checkers = namespace.ordered_checkers
        # Map each checker to whether its enabled or not.
        ordered_checkers.append((value, self.dest == 'enable'))

        namespace.ordered_checkers = ordered_checkers


class OrderedConfigAction(argparse.Action):
    """
    Action to store --analyzer-config and --checker-config values. These may
    come from many sources, including the CLI, the CodeChecker config file,
    the saargs file and the tidyargs file, or some other places.
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs != '*':
            raise ValueError("nargs must be '*' for backward compatibility "
                             "reasons!")
        super(OrderedConfigAction, self).__init__(option_strings, dest,
                                                  nargs, **kwargs)

    def __call__(self, parser, namespace, value, option_string=None):

        assert isinstance(value, list), \
               f"--analyzer-config or --checker-config value ({value}) is " \
               "not a list, but should be if nargs is not None!"

        if not hasattr(namespace, self.dest):
            setattr(namespace, self.dest, [])

        dest = getattr(namespace, self.dest)

        for flag in value:
            if flag in dest:
                dest.remove(flag)
            dest.append(flag)


def existing_abspath(path: str) -> str:
    """
    This function can be used at "type" argument of argparse.add_argument()
    function. It returns the absolute path of the given path if exists
    otherwise raises an argparse.ArgumentTypeError which constitutes in a
    graceful error message automatically by argparse module.
    """
    path = os.path.abspath(path)

    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f"File doesn't exist: {path}")

    return path


def analyzer_config(arg: str) -> AnalyzerConfig:
    """
    This function can be used at "type" argument of argparse.add_argument().
    It checks the format of --analyzer-config flag:
    <analyzer>:<option>=<value>
    These three things return as a tuple.
    """
    m = re.match(r"(?P<analyzer>.+):(?P<option>.+)=(?P<value>.+)", arg)

    if not m:
        raise argparse.ArgumentTypeError(
            f"Analyzer option in wrong format: {arg}, should be "
            "<analyzer>:<option>=<value>")

    return AnalyzerConfig(
        m.group("analyzer"), m.group("option"), m.group("value"))


def checker_config(arg: str) -> CheckerConfig:
    """
    This function can be used at "type" argument of argparse.add_argument().
    It checks the format of --checker-config flag:
    <analyzer>:<checker>:<option>=<value>
    These four things return as a tuple.
    """
    m = re.match(
        r"(?P<analyzer>.+):(?P<checker>.+):(?P<option>.+)=(?P<value>.+)", arg)

    if not m:
        raise argparse.ArgumentTypeError(
            f"Checker option in wrong format: {arg}, should be "
            "<analyzer>:<checker>:<option>=<value>")

    return CheckerConfig(
        m.group("analyzer"), m.group("checker"),
        m.group("option"), m.group("value"))


def analyzer_binary(arg: str) -> AnalyzerBinary:
    """
    This function can be used at "type" argument of argparse.add_argument().
    It checks the format of --analyzer_binary flag: <analyzer>:<path>
    """
    m = re.match(r"(?P<analyzer>.+):(?P<path>.+)", arg)

    if not m:
        raise argparse.ArgumentTypeError(
            f"Analyzer binary in wrong format: {arg}, should be "
            "<analyzer>:</path/to/bin/>")

    return AnalyzerBinary(m.group("analyzer"), m.group("path"))



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Helper for the configuration thrift api.
"""

from codechecker_api.Configuration_v6 import configurationService

from codechecker_client.thrift_call import ThriftClientCall
from .base import BaseClientHelper


class ThriftConfigHelper(BaseClientHelper):
    def __init__(self, protocol, host, port, uri, session_token=None):
        super().__init__(protocol, host, port, uri, session_token)

        self.client = configurationService.Client(self.protocol)

    # -----------------------------------------------------------------------
    @ThriftClientCall
    def getNotificationBannerText(self):
        pass



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import json
import logging
import os

from abc import ABCMeta, abstractmethod
from collections import defaultdict
import hashlib
from typing import Dict, Iterable, List, Optional

from codechecker_report_converter.report import Report, report_file
from codechecker_report_converter.report.hash import get_report_hash, HashType
from codechecker_report_converter.report.parser.base import AnalyzerInfo


LOG = logging.getLogger('report-converter')


class AnalyzerResultBase(metaclass=ABCMeta):
    """ Base class to transform analyzer result. """

    # Short name of the analyzer.
    TOOL_NAME: str = ''

    # Full name of the analyzer.
    NAME: str = ''

    # Link to the official analyzer website.
    URL: str = ''

    def transform(
        self,
        analyzer_result_file_paths: Iterable[str],
        output_dir_path: str,
        export_type: str,
        file_name: str = "{source_file}_{analyzer}_{file_hash}",
        metadata: Optional[Dict[str, str]] = None
    ) -> bool:
        """
        Converts the given analyzer result to the output directory in the given
        output type.
        """
        parser = report_file.get_parser(f".{export_type}")
        if not parser:
            LOG.error("The given output type '%s' is not supported!",
                      export_type)
            return False

        all_reports = []

        for file_path in analyzer_result_file_paths:
            file_path = os.path.abspath(file_path)
            reports = self.get_reports(file_path)
            if not reports:
                LOG.info("No '%s' results can be found in '%s'.",
                         self.TOOL_NAME, file_path)

            self._post_process_result(reports)

            for report in reports:
                report.analyzer_result_file_path = file_path

                if not report.checker_name:
                    report.checker_name = self.TOOL_NAME

            all_reports.extend(reports)

        self._write(
            all_reports, output_dir_path, parser, export_type, file_name)

        if metadata:
            self._save_metadata(metadata, output_dir_path)
        else:
            LOG.warning("Use '--meta' option to provide extra information "
                        "to the CodeChecker server such as analyzer version "
                        "and analysis command when storing the results to it. "
                        "For more information see the --help.")

        return bool(all_reports)

    @abstractmethod
    def get_reports(self, analyzer_result_file_path: str) -> List[Report]:
        """ Get reports from the given analyzer result. """
        raise NotImplementedError("Subclasses should implement this!")

    def _save_metadata(self, metadata, output_dir):
        """ Save metadata.json file to the output directory which will be used
        by CodeChecker.
        """
        meta_info = {
            "version": 2,
            "num_of_report_dir": 1,
            "tools": []
        }

        tool = {"name": self.TOOL_NAME}

        if "analyzer_version" in metadata:
            tool["version"] = metadata["analyzer_version"]

        if "analyzer_command" in metadata:
            tool["command"] = metadata["analyzer_command"]

        meta_info["tools"].append(tool)

        metadata_file = os.path.join(output_dir, 'metadata.json')
        with open(metadata_file, 'w',
                  encoding="utf-8", errors="ignore") as metafile:
            json.dump(meta_info, metafile)

    def _post_process_result(self, reports: List[Report]):
        """ Post process the parsed result.

        By default it will add report hashes and metada information.
        """
        for report in reports:
            self._add_report_hash(report)
            self._add_metadata(report)

    def _add_report_hash(self, report: Report):
        """ Generate report hash for the given plist data. """
        report.report_hash = get_report_hash(report, HashType.CONTEXT_FREE)

    def _add_metadata(self, report: Report):
        """ Add metada information to the given plist data. """
        report.analyzer_name = self.TOOL_NAME

    def _write(
        self,
        reports: List[Report],
        output_dir_path: str,
        parser,
        export_type: str,
        file_name: str
    ):
        """ Creates plist files from the parse result to the given output.

        It will generate a context free hash for each diagnostics.
        """
        output_dir = os.path.abspath(output_dir_path)

        file_to_report: Dict[str, List[Report]] = defaultdict(list)
        for report in reports:
            file_path = os.path.normpath(report.file.original_path)
            file_to_report[file_path].append(report)

        analyzer_info = AnalyzerInfo(name=self.TOOL_NAME)
        for file_path, file_reports in file_to_report.items():
            source_file = os.path.basename(file_path)
            file_hash = hashlib.md5(file_path.encode(errors='ignore')) \
                .hexdigest()

            out_file_name = file_name \
                .replace("{source_file}", source_file) \
                .replace("{analyzer}", self.TOOL_NAME) \
                .replace("{file_hash}", file_hash)
            out_file_name = f"{out_file_name}.{export_type}"
            out_file_path = os.path.join(output_dir, out_file_name)

            LOG.info("Create/modify plist file: '%s'.", out_file_path)
            LOG.debug(file_reports)

            data = parser.convert(file_reports, analyzer_info)
            parser.write(data, out_file_path)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handle Thrift requests for the product manager service.
"""


import os
import random

from sqlalchemy.sql.expression import and_

import codechecker_api_shared
from codechecker_api.ProductManagement_v6 import ttypes

from codechecker_common.logger import get_logger

from codechecker_server.profiler import timeit
from codechecker_web.shared import convert

from .. import permissions
from ..database.config_db_model import IDENTIFIER, Product, ProductPermission
from ..database.database import DBSession, SQLServer, conv, escape_like
from ..routing import is_valid_product_endpoint

from .thrift_enum_helper import confidentiality_enum, \
        confidentiality_str

LOG = get_logger('server')


class ThriftProductHandler:
    """
    Connect to database and handle thrift client requests.
    """

    def __init__(self,
                 server,
                 auth_session,
                 config_session,
                 routed_product,
                 package_version):

        self.__server = server
        self.__auth_session = auth_session
        self.__package_version = package_version
        self.__session = config_session
        self.__product = routed_product

        self.__permission_args = {
            'productID': routed_product.id if routed_product else None
        }

    def __require_permission(self, required, args=None):
        """
        Helper method to raise an UNAUTHORIZED exception if the user does not
        have any of the given permissions.
        """

        with DBSession(self.__session) as session:
            if args is None:
                args = dict(self.__permission_args)
                args['config_db_session'] = session

            if not any([permissions.require_permission(
                            perm, args, self.__auth_session)
                        for perm in required]):
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                    "You are not authorized to execute this action.")

            return True

    def __administrating(self, args):
        """ True if the current user can administrate the given product. """
        if permissions.require_permission(permissions.SUPERUSER, args,
                                          self.__auth_session):
            return True

        if permissions.require_permission(permissions.PRODUCT_ADMIN, args,
                                          self.__auth_session):
            return True

        return False

    def __get_product(self, session, product):
        """
        Retrieve the product connection object and create a Thrift Product
        object for the given product record in the database.
        """

        server_product = self.__server.get_product(product.endpoint)
        if not server_product:
            LOG.info("Product '%s' was found in the configuration "
                     "database but no database connection was "
                     "present. Mounting analysis run database...",
                     product.endpoint)
            self.__server.add_product(product)
            server_product = self.__server.get_product(product.endpoint)

        descr = convert.to_b64(product.description) \
            if product.description else None

        args = {'config_db_session': session,
                'productID': product.id}
        product_access = permissions.require_permission(
            permissions.PRODUCT_VIEW, args, self.__auth_session)

        admin_perm_name = permissions.PRODUCT_ADMIN.name
        admins = session.query(ProductPermission). \
            filter(and_(ProductPermission.permission == admin_perm_name,
                        ProductPermission.product_id == product.id)) \
            .all()

        connected = server_product.db_status ==\
            codechecker_api_shared.ttypes.DBStatus.OK

        latest_storage_date = str(product.latest_storage_date) \
            if product.latest_storage_date else None

        if product.confidentiality is None:
            confidentiality = ttypes.Confidentiality.CONFIDENTIAL
        else:
            confidentiality = \
                    confidentiality_enum(product.confidentiality)

        report_limit = product.report_limit

        return server_product, ttypes.Product(
            id=product.id,
            endpoint=product.endpoint,
            displayedName_b64=convert.to_b64(product.display_name),
            description_b64=descr,
            runCount=product.num_of_runs,
            latestStoreToProduct=latest_storage_date,
            connected=connected,
            accessible=product_access,
            administrating=self.__administrating(args),
            databaseStatus=server_product.db_status,
            admins=[admin.name for admin in admins],
            confidentiality=confidentiality,
            reportLimit=report_limit)

    @timeit
    def getPackageVersion(self):
        return self.__package_version

    @timeit
    def isAdministratorOfAnyProduct(self):
        with DBSession(self.__session) as session:
            prods = session.query(Product).all()

            for prod in prods:
                args = {'config_db_session': session,
                        'productID': prod.id}
                if permissions.require_permission(
                        permissions.PRODUCT_ADMIN,
                        args, self.__auth_session):
                    return True

            return False

    @timeit
    def getProducts(self, product_endpoint_filter, product_name_filter):
        """
        Get the list of products configured on the server.
        """

        result = []

        with DBSession(self.__session) as session:
            prods = session.query(Product)

            num_all_products = prods.count()  # prods get filtered later.
            if num_all_products < self.__server.num_products:
                # It can happen that a product gets removed from the
                # configuration database from a different server that uses the
                # same configuration database. In this case, the product is
                # no longer valid, yet the current server keeps a connection
                # object up.
                LOG.info("%d products were removed but server is still "
                         "connected to them. Disconnecting these...",
                         self.__server.num_products - num_all_products)

                all_products = session.query(Product).all()
                self.__server.remove_products_except([prod.endpoint for prod
                                                      in all_products])

            if product_endpoint_filter:
                prods = prods.filter(Product.endpoint.ilike(
                    conv(escape_like(product_endpoint_filter, '\\')),
                    escape='\\'))

            if product_name_filter:
                prods = prods.filter(Product.display_name.ilike(
                    conv(escape_like(product_name_filter, '\\')),
                    escape='\\'))

            prods = prods.all()
            for prod in prods:
                _, ret = self.__get_product(session, prod)
                result.append(ret)

            return result

    @timeit
    def getCurrentProduct(self):
        """
        Return information about the current product.

        The request MUST be routed as /product-name/ProductService!
        """

        if not self.__product:
            msg = "Requested current product from ProductService but the " \
                  "request came through the main endpoint."
            LOG.error(msg)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.IOERROR,
                msg)

        with DBSession(self.__session) as session:
            prod = session.query(Product).get(self.__product.id)

            if not prod:
                msg = "The product requested has been disconnected from the " \
                      "server."
                LOG.error(msg)
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.IOERROR,
                    msg)

            _, ret = self.__get_product(session, prod)
            LOG.debug(ret)
            return ret

    @timeit
    def getProductConfiguration(self, product_id):
        """
        Get the product configuration --- WITHOUT THE DB PASSWORD --- of the
        given product.
        """

        with DBSession(self.__session) as session:
            product = session.query(Product).get(product_id)
            if product is None:
                msg = "Product with ID {0} does not exist!".format(product_id)
                LOG.error(msg)

                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.DATABASE, msg)

            # Put together the database connection's descriptor.
            args = SQLServer.connection_string_to_args(product.connection)
            if args['postgresql']:
                db_engine = 'postgresql'
                db_host = args['dbaddress']
                db_port = args['dbport']
                db_user = args['dbusername']
                db_name = args['dbname']
            else:
                db_engine = 'sqlite'
                db_host = ""
                db_port = 0
                db_user = ""
                db_name = args['sqlite']

            dbc = ttypes.DatabaseConnection(
                engine=db_engine,
                host=db_host,
                port=db_port,
                username_b64=convert.to_b64(db_user),
                # DO NOT TRANSPORT PASSWORD BACK TO THE USER!
                database=db_name)

            # Put together the product configuration.
            descr = convert.to_b64(product.description) \
                if product.description else None

            is_review_status_change_disabled = \
                product.is_review_status_change_disabled

            if product.confidentiality is None:
                confidentiality = ttypes.Confidentiality.CONFIDENTIAL
            else:
                confidentiality = \
                        confidentiality_enum(product.confidentiality)

            prod = ttypes.ProductConfiguration(
                id=product.id,
                endpoint=product.endpoint,
                displayedName_b64=convert.to_b64(product.display_name),
                description_b64=descr,
                connection=dbc,
                runLimit=product.run_limit,
                reportLimit=product.report_limit,
                isReviewStatusChangeDisabled=is_review_status_change_disabled,
                confidentiality=confidentiality)

            return prod

    @timeit
    def addProduct(self, product):
        """
        Add the given product to the products configured by the server.
        """
        self.__require_permission([permissions.SUPERUSER])

        session = None
        LOG.info("User requested add product '%s'", product.endpoint)

        if not is_valid_product_endpoint(product.endpoint):
            msg = "The specified endpoint is invalid."
            LOG.error(msg)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                msg)

        dbc = product.connection
        if not dbc:
            msg = "Product cannot be added without a database configuration!"
            LOG.error(msg)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                msg)

        if self.__server.get_product(product.endpoint):
            msg = "A product endpoint '/{0}' is already configured!" \
                .format(product.endpoint)
            LOG.error(msg)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                msg)

        # Some values come encoded as Base64, decode these.
        displayed_name = convert.from_b64(product.displayedName_b64) \
            if product.displayedName_b64 else product.endpoint
        description = convert.from_b64(product.description_b64) \
            if product.description_b64 else None

        if dbc.engine == 'sqlite' and not os.path.isabs(dbc.database):
            # Transform the database relative path to be under the
            # server's config directory.
            dbc.database = os.path.join(self.__server.config_directory,
                                        dbc.database)

        # Transform arguments into a database connection string.
        if dbc.engine == 'postgresql':
            dbuser = "codechecker"
            dbpass = ""
            if dbc.username_b64 and dbc.username_b64 != '':
                dbuser = convert.from_b64(dbc.username_b64)
            if dbc.password_b64 and dbc.password_b64 != '':
                dbpass = convert.from_b64(dbc.password_b64)

            conn_str_args = {'postgresql': True,
                             'sqlite': False,
                             'dbaddress': dbc.host,
                             'dbport': dbc.port,
                             'dbusername': dbuser,
                             'dbpassword': dbpass,
                             'dbname': dbc.database}
        elif dbc.engine == 'sqlite':
            conn_str_args = {'postgresql': False,
                             'sqlite': dbc.database}
        else:
            msg = "Database engine '{0}' unknown!".format(dbc.engine)
            LOG.error(msg)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                msg)

        conn_str = SQLServer \
            .from_cmdline_args(conn_str_args, product.endpoint, IDENTIFIER,
                               None, False, None) \
            .get_connection_string()

        is_rws_change_disabled = product.isReviewStatusChangeDisabled

        confidentiality = confidentiality_str(product.confidentiality)

        # Create the product's entity in the database.
        with DBSession(self.__session) as session:
            orm_prod = Product(
                endpoint=product.endpoint,
                conn_str=conn_str,
                name=displayed_name,
                description=description,
                run_limit=product.runLimit,
                report_limit=product.reportLimit,
                is_review_status_change_disabled=is_rws_change_disabled,
                confidentiality=confidentiality)

            LOG.debug("Attempting database connection to new product...")

            # Connect and create the database schema.
            self.__server.add_product(orm_prod, init_db=True)
            connection_wrapper = self.__server.get_product(product.endpoint)
            if connection_wrapper.last_connection_failure:
                msg = "The configured connection for '/{0}' failed: {1}" \
                    .format(product.endpoint,
                            connection_wrapper.last_connection_failure)
                LOG.error(msg)

                self.__server.remove_product(product.endpoint)

                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.IOERROR, msg)

            LOG.debug("Product database successfully connected to.")
            session.add(orm_prod)
            session.flush()

            # Create the default permissions for the product
            permissions.initialise_defaults('PRODUCT', {
                'config_db_session': session,
                'productID': orm_prod.id
            })
            session.commit()
            LOG.debug("Product configuration added to database successfully.")

            # The orm_prod object above is not bound to the database as it
            # was just created. We use the actual database-backed configuration
            # entry to handle connections, so a "reconnect" is issued here.
            self.__server.remove_product(product.endpoint)

            orm_prod = session.query(Product) \
                .filter(Product.endpoint == product.endpoint).one()
            self.__server.add_product(orm_prod)

            LOG.debug("Product database connected and ready to serve.")
            return True

    @timeit
    def editProduct(self, product_id, new_config):
        """
        Edit the given product's properties to the one specified by
        new_configuration.
        """
        with DBSession(self.__session) as session:
            product = session.query(Product).get(product_id)
            if product is None:
                msg = "Product with ID {0} does not exist!".format(product_id)
                LOG.error(msg)
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.DATABASE, msg)

            # Editing the metadata of the product, such as display name and
            # description is available for product admins.

            # Because this query doesn't come through a product endpoint,
            # __init__ sets the value in the extra_args to None.
            self.__permission_args['productID'] = product.id
            self.__require_permission([permissions.PRODUCT_ADMIN])

            LOG.info("User requested edit product '%s'",
                     product.endpoint)

            dbc = new_config.connection
            if not dbc:
                msg = "Product's database configuration cannot be removed!"
                LOG.error(msg)
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.GENERAL, msg)

            if new_config.endpoint != product.endpoint:
                if not is_valid_product_endpoint(new_config.endpoint):
                    msg = "The endpoint to move the product to is invalid."
                    LOG.error(msg)
                    raise codechecker_api_shared.ttypes.RequestFailed(
                        codechecker_api_shared.ttypes.ErrorCode.GENERAL, msg)

                if self.__server.get_product(new_config.endpoint):
                    msg = f"A product endpoint '/{product.endpoint}' is " \
                          f"already configured!"
                    LOG.error(msg)
                    raise codechecker_api_shared.ttypes.RequestFailed(
                        codechecker_api_shared.ttypes.ErrorCode.GENERAL, msg)

                LOG.info("User renamed product '%s' to '%s'",
                         product.endpoint, new_config.endpoint)

            # Some values come encoded as Base64, decode these.
            displayed_name = convert.from_b64(new_config.displayedName_b64) \
                if new_config.displayedName_b64 \
                else new_config.endpoint
            description = convert.from_b64(new_config.description_b64) \
                if new_config.description_b64 else None

            confidentiality = confidentiality_str(new_config.confidentiality)

            if dbc.engine == 'sqlite' and not os.path.isabs(dbc.database):
                # Transform the database relative path to be under the
                # server's config directory.
                dbc.database = os.path.join(self.__server.config_directory,
                                            dbc.database)

            # Transform arguments into a database connection string.
            if dbc.engine == 'postgresql':
                dbuser = "codechecker"
                if dbc.username_b64 and dbc.username_b64 != '':
                    dbuser = convert.from_b64(dbc.username_b64)

                old_connection_args = SQLServer.connection_string_to_args(
                        product.connection)
                if dbc.password_b64 and dbc.password_b64 != '':
                    dbpass = convert.from_b64(dbc.password_b64)
                elif 'dbpassword' in old_connection_args:
                    # The password was not changed. Retrieving from old
                    # configuration -- if the old configuration contained such!
                    dbpass = old_connection_args['dbpassword']
                else:
                    dbpass = None

                conn_str_args = {'postgresql': True,
                                 'sqlite': False,
                                 'dbaddress': dbc.host,
                                 'dbport': dbc.port,
                                 'dbusername': dbuser,
                                 'dbpassword': dbpass,
                                 'dbname': dbc.database}
            elif dbc.engine == 'sqlite':
                conn_str_args = {'postgresql': False,
                                 'sqlite': dbc.database}
            else:
                msg = "Database engine '{0}' unknown!".format(dbc.engine)
                LOG.error(msg)
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                    msg)

            conn_str = SQLServer \
                .from_cmdline_args(conn_str_args, product.endpoint,
                                   IDENTIFIER, None, False, None) \
                .get_connection_string()

            # If endpoint or database arguments change, the product
            # configuration has changed so severely, that it needs
            # to be reconnected.
            product_needs_reconnect = \
                product.endpoint != new_config.endpoint or \
                product.connection != conn_str
            old_endpoint = product.endpoint

            if product_needs_reconnect:
                # Changing values that alter connection-specific data
                # should only be available for superusers!
                self.__require_permission([permissions.SUPERUSER])

                # Test if the new database settings are correct or not.
                dummy_endpoint = new_config.endpoint + '_' + ''.join(
                    random.sample(new_config.endpoint,
                                  min(len(new_config.endpoint), 5)))
                dummy_prod = Product(
                    endpoint=dummy_endpoint,
                    conn_str=conn_str,
                    name=displayed_name,
                    description=description)

                LOG.debug("Attempting database connection with new "
                          "settings...")

                # Connect and create the database schema.
                self.__server.add_product(dummy_prod)
                LOG.debug("Product database successfully connected to.")

                connection_wrapper = self.__server.get_product(dummy_endpoint)
                if connection_wrapper.last_connection_failure:
                    msg = "The configured connection for '/{0}' failed: {1}" \
                        .format(new_config.endpoint,
                                connection_wrapper.last_connection_failure)
                    LOG.error(msg)

                    self.__server.remove_product(dummy_endpoint)

                    raise codechecker_api_shared.ttypes.RequestFailed(
                        codechecker_api_shared.ttypes.ErrorCode.IOERROR, msg)

                # The orm_prod object above is not bound to the database as it
                # was just created. We use the actual database-backed
                # configuration entry to handle connections, so a "reconnect"
                # is issued later.
                self.__server.remove_product(dummy_endpoint)

            # Update the settings in the database.
            product.endpoint = new_config.endpoint
            product.run_limit = new_config.runLimit
            product.report_limit = new_config.reportLimit
            product.is_review_status_change_disabled = \
                new_config.isReviewStatusChangeDisabled
            product.connection = conn_str
            product.display_name = displayed_name
            product.description = description
            product.confidentiality = confidentiality

            session.commit()
            LOG.info("Product configuration edited and saved successfully.")

            if product_needs_reconnect:
                product = session.query(Product).get(product_id)
                LOG.info("Product change requires database reconnection...")

                LOG.debug("Disconnecting...")
                try:
                    # Because of the process pool it is possible that in the
                    # local cache of the current process the product with the
                    # old endpoint is not found and it will raise an exception.
                    self.__server.remove_product(old_endpoint)
                except ValueError:
                    pass

                LOG.debug("Connecting new settings...")
                self.__server.add_product(product)

                LOG.info("Product reconnected successfully.")

            return True

    @timeit
    def removeProduct(self, product_id):
        """
        Disconnect the product specified by the ID from the server.
        """
        self.__require_permission([permissions.SUPERUSER])

        with DBSession(self.__session) as session:
            product = session.query(Product).get(product_id)
            if product is None:
                msg = "Product with ID {0} does not exist!".format(product_id)
                LOG.error(msg)

                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.DATABASE, msg)

            LOG.info("User requested to remove product '%s'", product.endpoint)
            self.__server.remove_product(product.endpoint)

            session.delete(product)
            session.commit()
            return True



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Multiprocessing compatibility module.
"""
import sys

# pylint: disable=no-name-in-module,unused-import
if sys.platform in ["darwin", "win32"]:
    from multiprocess import Pool  # type: ignore
    from multiprocess import cpu_count
else:
    from concurrent.futures import ProcessPoolExecutor as Pool  # type: ignore
    from multiprocessing import cpu_count



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import re
from typing import Iterator, Optional, Tuple

from codechecker_report_converter.report import Report

from ..parser import SANParser

LOG = logging.getLogger('report-converter')


class Parser(SANParser):
    """ Parser for Clang AddressSanitizer console outputs. """

    checker_name = "AddressSanitizer"

    # Regex for parsing AddressSanitizer output message.
    line_re = re.compile(
            # Error code
            r'==(?P<code>\d+)==(ERROR|WARNING): AddressSanitizer: '
            # Checker message.
            r'(?P<message>[\S \t]+)')

    # Sources: https://github.com/llvm/llvm-project/blob/main/compiler-rt/lib/
    #   sanitizer_common/sanitizer_allocator_report.cpp
    #   asan/asan_errors.h
    #   asan/asan_errors.cpp
    #    -> both inside ErrorGeneric::ErrorGeneric() and
    #                   ErrorStringFunction.*::Print()
    check_msg_pairs = {
        "double-free": re.compile(
            "attempting double-free on .* in thread .*:"),
        "new-delete-type-mismatch": re.compile(
            "new-delete-type-mismatch on .* in thread .*:"),
        "bad-free": re.compile(
            "attempting free on address which was not malloc()-ed: .* in "
            "thread .*"),
        "alloc-dealloc-mismatch": re.compile(
            "alloc-dealloc-mismatch (.* vs .*) on .*"),
        "bad-malloc_usable_size": re.compile(
            "attempting to call malloc_usable_size() for pointer which is not "
            "not owned: .*"),
        "bad-__sanitizer_get_allocated_size": re.compile(
            "attempting to call __sanitizer_get_allocated_size() for pointer "
            "which is not owned: .*"),
        "calloc-overflow": re.compile(
            "calloc parameters overflow: count * size (.* * .*) cannot be "
            "represented in type size_t (thread .*)"),
        "reallocarray-overflow": re.compile(
            "reallocarray parameters overflow: count * size (.* * .*) cannot "
            "be represented in type size_t (thread .*)"),
        "pvalloc-overflow": re.compile(
            "pvalloc parameters overflow: size 0x.* rounded up to system page "
            "size 0x.* cannot be represented in type size_t (thread .*)"),
        "invalid-allocation-alignment": re.compile(
            "invalid allocation alignment: .*, alignment must be a power of "
            "two (thread .*)"),
        "invalid-aligned-alloc-alignment": re.compile(
            "invalid alignment requested in aligned_alloc: .* the requested "
            "size 0x.* must be a multiple of alignment (thread .*)"),
        "invalid-posix-memalign-alignment": re.compile(
            "invalid alignment requested in posix_memalign: .*, alignment "
            "must be a power of two and a multiple of sizeof(void*) == .* "
            "(thread .*)"),
        "allocation-size-too-big": re.compile(
            "requested allocation size 0x.* (0x.* after adjustments for "
            "alignment, red zones etc.) exceeds maximum supported size of "
            "0x.* (thread .*)"),
        "rss-limit-exceeded": re.compile(
            "specified RSS limit exceeded, currently set to "
            "soft_rss_limit_mb=.*"),
        "out-of-memory": re.compile(
            "allocator is trying to allocate 0x.* bytes"),
        # This is not a check name in LLVM, but is the name of the class that
        # emits these errors.
        "string-function-memory-ranges-overlap": re.compile(
            ".*: memory ranges .* and .* overlap"),
        "negative-size-param": re.compile(
            "negative-size-param: (size=.*)"),
        "bad-__sanitizer_annotate_contiguous_container": re.compile(
            "bad parameters to __sanitizer_annotate_contiguous_container:"),
        "bad-__sanitizer_annotate_double_ended_contiguous_container":
        re.compile(
            "bad parameters to "
            "__sanitizer_annotate_double_ended_contiguous_container:"),
        "odr-violation": re.compile("odr-violation (.*):"),
        "invalid-pointer-pair": re.compile("invalid-pointer-pair: .* .*"),
        # The name of the class that handles this case is called GenericError,
        # so this is as good of a name as we can come up with.
        # heap-use-after-free and reports like that are emitted from this.
        "generic-error": re.compile(".* on address .* at pc .* bp .* sp .*")
    }

    def deduce_checker_name(self, message: str) -> str:
        for check, pattern in self.check_msg_pairs.items():
            if pattern.search(message):
                return f"{self.checker_name}.{check}"
        return self.checker_name

    def parse_sanitizer_message(
        self,
        it: Iterator[str],
        line: str
    ) -> Tuple[Optional[Report], str]:
        report, line = super().parse_sanitizer_message(it, line)

        if not report:
            return report, line

        report.checker_name = self.deduce_checker_name(report.message)

        return report, line
